!MNH_LIC Copyright 1994-2021 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
MODULE MODI_ICE4_SLOW
INTERFACE
SUBROUTINE ICE4_SLOW(LDSOFT, PCOMPUTE, PRHODREF, PT,&
                     &PSSI, PLVFACT, PLSFACT, &
                     &PRVT, PRCT, PRIT, PRST, PRGT,&
                     &PLBDAS, PLBDAG,&
                     &PAI, PCJ, PHLI_HCF, PHLI_HRI,&
                     &PRCHONI, PRVDEPS, PRIAGGS, PRIAUTS, PRVDEPG, &
                     &PA_TH, PA_RV, PA_RC, PA_RI, PA_RS, PA_RG)
IMPLICIT NONE
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
REAL, DIMENSION(:),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PSSI     ! Supersaturation over ice
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PRVT
REAL, DIMENSION(:),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRIT     ! Pristine ice m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRST     ! Snow/aggregate m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRGT     ! Graupel/hail m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAG   ! Slope parameter of the graupel   distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PAI      ! Thermodynamical function
REAL, DIMENSION(:),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(:),       INTENT(IN)    :: PHLI_HCF !
REAL, DIMENSION(:),       INTENT(IN)    :: PHLI_HRI !
REAL, DIMENSION(:),       INTENT(INOUT) :: PRCHONI  ! Homogeneous nucleation
REAL, DIMENSION(:),       INTENT(INOUT) :: PRVDEPS  ! Deposition on r_s
REAL, DIMENSION(:),       INTENT(INOUT) :: PRIAGGS  ! Aggregation on r_s
REAL, DIMENSION(:),       INTENT(INOUT) :: PRIAUTS  ! Autoconversion of r_i for r_s production
REAL, DIMENSION(:),       INTENT(INOUT) :: PRVDEPG  ! Deposition on r_g
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RV
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RI
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RS
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RG
END SUBROUTINE ICE4_SLOW
END INTERFACE
END MODULE MODI_ICE4_SLOW
SUBROUTINE ICE4_SLOW(LDSOFT, PCOMPUTE, PRHODREF, PT, &
                     &PSSI, PLVFACT, PLSFACT, &
                     &PRVT, PRCT, PRIT, PRST, PRGT, &
                     &PLBDAS, PLBDAG, &
                     &PAI, PCJ, PHLI_HCF, PHLI_HRI,&
                     &PRCHONI, PRVDEPS, PRIAGGS, PRIAUTS, PRVDEPG, &
                     &PA_TH, PA_RV, PA_RC, PA_RI, PA_RS, PA_RG)
!!
!!**  PURPOSE
!!    -------
!!      Computes the slow process
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XTT
USE MODD_RAIN_ICE_DESCR, ONLY: XCEXVT,XRTMIN
USE MODD_RAIN_ICE_PARAM, ONLY: X0DEPG,X0DEPS,X1DEPG,X1DEPS,XACRIAUTI,XALPHA3,XBCRIAUTI,XBETA3,XCOLEXIS,XCRIAUTI, &
                               XEX0DEPG,XEX0DEPS,XEX1DEPG,XEX1DEPS,XEXIAGGS,XFIAGGS,XHON,XTEXAUTI,XTIMAUTI
!
#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
USE MODE_MPPDB
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
REAL, DIMENSION(:),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PSSI     ! Supersaturation over ice
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PRVT
REAL, DIMENSION(:),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRIT     ! Pristine ice m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRST     ! Snow/aggregate m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRGT     ! Graupel/hail m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAG   ! Slope parameter of the graupel   distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PAI      ! Thermodynamical function
REAL, DIMENSION(:),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(:),       INTENT(IN)    :: PHLI_HCF !
REAL, DIMENSION(:),       INTENT(IN)    :: PHLI_HRI !
REAL, DIMENSION(:),       INTENT(INOUT) :: PRCHONI  ! Homogeneous nucleation
REAL, DIMENSION(:),       INTENT(INOUT) :: PRVDEPS  ! Deposition on r_s
REAL, DIMENSION(:),       INTENT(INOUT) :: PRIAGGS  ! Aggregation on r_s
REAL, DIMENSION(:),       INTENT(INOUT) :: PRIAUTS  ! Autoconversion of r_i for r_s production
REAL, DIMENSION(:),       INTENT(INOUT) :: PRVDEPG  ! Deposition on r_g
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RV
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RI
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RS
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RG
!
!*       0.2  declaration of local variables
!
INTEGER                         :: ISIZE
INTEGER                         :: JL
REAL, DIMENSION(:), allocatable :: ZMASK
REAL, DIMENSION(:), allocatable :: ZCRIAUTI
REAL                            :: ZTIMAUTIC
!
!$acc data present(PCOMPUTE,PRHODREF,PT,PSSI,PLVFACT,PLSFACT, &
!$acc&                PRVT,PRCT,PRIT,PRST,PRGT,PLBDAS,PLBDAG,PAI,PCJ,          &
!$acc&                PHLI_HCF, PHLI_HRI, PRCHONI,PRVDEPS,PRIAGGS,PRIAUTS,PRVDEPG, &
!$acc&                PA_TH,PA_RV,PA_RC,PA_RI,PA_RS,PA_RG                      )
!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PCOMPUTE,"ICE4_SLOW beg:PCOMPUTE")
  CALL MPPDB_CHECK(PRHODREF,"ICE4_SLOW beg:PRHODREF")
  CALL MPPDB_CHECK(PT,"ICE4_SLOW beg:PT")
  CALL MPPDB_CHECK(PSSI,"ICE4_SLOW beg:PSSI")
  CALL MPPDB_CHECK(PLVFACT,"ICE4_SLOW beg:PLVFACT")
  CALL MPPDB_CHECK(PLSFACT,"ICE4_SLOW beg:PLSFACT")
  CALL MPPDB_CHECK(PRVT,"ICE4_SLOW beg:PRVT")
  CALL MPPDB_CHECK(PRCT,"ICE4_SLOW beg:PRCT")
  CALL MPPDB_CHECK(PRIT,"ICE4_SLOW beg:PRIT")
  CALL MPPDB_CHECK(PRST,"ICE4_SLOW beg:PRST")
  CALL MPPDB_CHECK(PRGT,"ICE4_SLOW beg:PRGT")
  CALL MPPDB_CHECK(PLBDAS,"ICE4_SLOW beg:PLBDAS")
  CALL MPPDB_CHECK(PLBDAG,"ICE4_SLOW beg:PLBDAG")
  CALL MPPDB_CHECK(PAI,"ICE4_SLOW beg:PAI")
  CALL MPPDB_CHECK(PCJ,"ICE4_SLOW beg:PCJ")
  CALL MPPDB_CHECK(PHLI_HCF,"ICE4_SLOW beg:PHLI_HCF")
  CALL MPPDB_CHECK(PHLI_HRI,"ICE4_SLOW beg:PHLI_HRI")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRCHONI,"ICE4_SLOW beg:PRCHONI")
  CALL MPPDB_CHECK(PRVDEPS,"ICE4_SLOW beg:PRVDEPS")
  CALL MPPDB_CHECK(PRIAGGS,"ICE4_SLOW beg:PRIAGGS")
  CALL MPPDB_CHECK(PRIAUTS,"ICE4_SLOW beg:PRIAUTS")
  CALL MPPDB_CHECK(PRVDEPG,"ICE4_SLOW beg:PRVDEPG")
  CALL MPPDB_CHECK(PA_TH,"ICE4_SLOW beg:PA_TH")
  CALL MPPDB_CHECK(PA_RV,"ICE4_SLOW beg:PA_RV")
  CALL MPPDB_CHECK(PA_RC,"ICE4_SLOW beg:PA_RC")
  CALL MPPDB_CHECK(PA_RI,"ICE4_SLOW beg:PA_RI")
  CALL MPPDB_CHECK(PA_RS,"ICE4_SLOW beg:PA_RS")
  CALL MPPDB_CHECK(PA_RG,"ICE4_SLOW beg:PA_RG")
END IF

ISIZE = Size( PRHODREF )

allocate( zmask   (size( prhodref ) ) )
allocate( zcriauti(size( prhodref ) ) )

!$acc data create(ZCRIAUTI,ZMASK)

!
!*       3.2     compute the homogeneous nucleation source: RCHONI
!
!$acc kernels
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., PT(JL)-(XTT-35.0))) * & ! PT(:)<XTT-35.0
           &MAX(0., -SIGN(1., XRTMIN(2)-PRCT(JL))) * & ! PRCT(:)>XRTMIN(2)
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRCHONI(JL) = PRCHONI(JL) * ZMASK(JL)
  ENDDO
ELSE
  PRCHONI(:) = 0.
  WHERE(ZMASK(:)==1.)
    PRCHONI(:) = MIN(1000.,XHON*PRHODREF(:)*PRCT(:)       &
#ifndef MNH_BITREP
                                 *EXP( XALPHA3*(PT(:)-XTT)-XBETA3 ) )
#else
                                 *BR_EXP( XALPHA3*(PT(:)-XTT)-XBETA3 ) )
#endif
  ENDWHERE
ENDIF
DO JL=1, ISIZE
  PA_RI(JL) = PA_RI(JL) + PRCHONI(JL)
  PA_RC(JL) = PA_RC(JL) - PRCHONI(JL)
  PA_TH(JL) = PA_TH(JL) + PRCHONI(JL)*(PLSFACT(JL)-PLVFACT(JL))
ENDDO
!
!*       3.4    compute the deposition, aggregation and autoconversion sources
!
!
!*       3.4.2  compute the riming-conversion of r_c for r_i production: RCAUTI
!
!  ZZW(:) = 0.0
!  ZTIMAUTIC = SQRT( XTIMAUTI*XTIMAUTC )
!  WHERE ( (PRCT(:)>0.0) .AND. (PRIT(:)>0.0) .AND. (PRCS(:)>0.0) )
!    ZZW(:) = MIN( PRCS(:),ZTIMAUTIC * MAX( SQRT( PRIT(:)*PRCT(:) ),0.0 ) )
!    PRIS(:) = PRIS(:) + ZZW(:)
!    PRCS(:) = PRCS(:) - ZZW(:)
!    PTHS(:) = PTHS(:) + ZZW(:)*(PLSFACT(:)-PLVFACT(:)) ! f(L_f*(RCAUTI))
!  END WHERE
!
!*       3.4.3  compute the deposition on r_s: RVDEPS
!
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(1)-PRVT(JL))) * & !PRVT(:)>XRTMIN(1)
           &MAX(0., -SIGN(1., XRTMIN(5)-PRST(JL))) * & !PRST(:)>XRTMIN(5)
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRVDEPS(JL)=PRVDEPS(JL)*ZMASK(JL)
  ENDDO
ELSE
  PRVDEPS(:) = 0.
  WHERE(ZMASK(:)==1.)
    PRVDEPS(:) = ( PSSI(:)/(PRHODREF(:)*PAI(:)) ) *                               &
#ifndef MNH_BITREP
                 ( X0DEPS*PLBDAS(:)**XEX0DEPS + X1DEPS*PCJ(:)*PLBDAS(:)**XEX1DEPS )
#else
                 ( X0DEPS*BR_POW(PLBDAS(:),XEX0DEPS) + X1DEPS*PCJ(:)*BR_POW(PLBDAS(:),XEX1DEPS) )
#endif
  END WHERE
ENDIF
DO JL=1, ISIZE
  PA_RS(JL) = PA_RS(JL) + PRVDEPS(JL)
  PA_RV(JL) = PA_RV(JL) - PRVDEPS(JL)
  PA_TH(JL) = PA_TH(JL) + PRVDEPS(JL)*PLSFACT(JL)
ENDDO
!
!*       3.4.4  compute the aggregation on r_s: RIAGGS
!
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(4)-PRIT(JL))) * & ! PRIT(:)>XRTMIN(4)
           &MAX(0., -SIGN(1., XRTMIN(5)-PRST(JL))) * & ! PRST(:)>XRTMIN(5)
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRIAGGS(JL)=PRIAGGS(JL) * ZMASK(JL)
  ENDDO
ELSE
  PRIAGGS(:) = 0.
  WHERE(ZMASK(:)==1)
#ifndef MNH_BITREP
    PRIAGGS(:) = XFIAGGS * EXP( XCOLEXIS*(PT(:)-XTT) ) &
                         * PRIT(:)                      &
                         * PLBDAS(:)**XEXIAGGS          &
                         * PRHODREF(:)**(-XCEXVT)
#else
    PRIAGGS(:) = XFIAGGS * BR_EXP( XCOLEXIS*(PT(:)-XTT) ) &
                         * PRIT(:)                        &
                         * BR_POW(PLBDAS(:),XEXIAGGS)     &
                         * BR_POW(PRHODREF(:),-XCEXVT)
#endif
  END WHERE
ENDIF
DO JL=1, ISIZE
  PA_RS(JL) = PA_RS(JL) + PRIAGGS(JL)
  PA_RI(JL) = PA_RI(JL) - PRIAGGS(JL)
ENDDO
!
!*       3.4.5  compute the autoconversion of r_i for r_s production: RIAUTS
!
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(4)-PHLI_HRI(JL))) * & ! PHLI_HRI(:)>XRTMIN(4)
           &MAX(0., -SIGN(1., 1.E-20-PHLI_HCF(JL))) * & ! PHLI_HCF(:) .GT. 0.
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRIAUTS(JL) = PRIAUTS(JL) * ZMASK(JL)
  ENDDO
ELSE
  PRIAUTS(:) = 0.
  !ZCRIAUTI(:)=MIN(XCRIAUTI,10**(0.06*(PT(:)-XTT)-3.5))
#ifndef MNH_BITREP
  ZCRIAUTI(:)=MIN(XCRIAUTI,10**(XACRIAUTI*(PT(:)-XTT)+XBCRIAUTI))
  WHERE(ZMASK(:)==1.)
    PRIAUTS(:) = XTIMAUTI * EXP( XTEXAUTI*(PT(:)-XTT) ) &
                          * MAX( PHLI_HRI(:)/PHLI_HCF(:)-ZCRIAUTI(:),0.0 )
    PRIAUTS(:) = PHLI_HCF(:)*PRIAUTS(:)
  END WHERE
#else
  ZCRIAUTI(:)=MIN(XCRIAUTI,BR_POW(10.,XACRIAUTI*(PT(:)-XTT)+XBCRIAUTI))
  WHERE(ZMASK(:)==1.)
    PRIAUTS(:) = XTIMAUTI * BR_EXP( XTEXAUTI*(PT(:)-XTT) ) &
                          * MAX( PRIT(:)-ZCRIAUTI(:),0.0 )
  END WHERE
#endif
ENDIF
DO JL=1, ISIZE
  PA_RS(JL) = PA_RS(JL) + PRIAUTS(JL)
  PA_RI(JL) = PA_RI(JL) - PRIAUTS(JL)
ENDDO
!
!*       3.4.6  compute the deposition on r_g: RVDEPG
!
!
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(1)-PRVT(JL))) * & ! PRVT(:)>XRTMIN(1)
           &MAX(0., -SIGN(1., XRTMIN(6)-PRGT(JL))) * & ! PRGT(:)>XRTMIN(6)
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRVDEPG(JL) = PRVDEPG(JL) * ZMASK(JL)
  ENDDO
ELSE
  PRVDEPG(:) = 0.
  WHERE(ZMASK(:)==1.)
    PRVDEPG(:) = ( PSSI(:)/(PRHODREF(:)*PAI(:)) ) *                               &
#ifndef MNH_BITREP
                 ( X0DEPG*PLBDAG(:)**XEX0DEPG + X1DEPG*PCJ(:)*PLBDAG(:)**XEX1DEPG )
#else
                 ( X0DEPG*BR_POW(PLBDAG(:),XEX0DEPG) + X1DEPG*PCJ(:)*BR_POW(PLBDAG(:),XEX1DEPG) )
#endif
  END WHERE
ENDIF
DO JL=1, ISIZE
  PA_RG(JL) = PA_RG(JL) + PRVDEPG(JL)
  PA_RV(JL) = PA_RV(JL) - PRVDEPG(JL)
  PA_TH(JL) = PA_TH(JL) + PRVDEPG(JL)*PLSFACT(JL)
ENDDO
!$acc end kernels
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRCHONI,"ICE4_SLOW end:PRCHONI")
  CALL MPPDB_CHECK(PRVDEPS,"ICE4_SLOW end:PRVDEPS")
  CALL MPPDB_CHECK(PRIAGGS,"ICE4_SLOW end:PRIAGGS")
  CALL MPPDB_CHECK(PRIAUTS,"ICE4_SLOW end:PRIAUTS")
  CALL MPPDB_CHECK(PRVDEPG,"ICE4_SLOW end:PRVDEPG")
  CALL MPPDB_CHECK(PA_TH,"ICE4_SLOW end:PA_TH")
  CALL MPPDB_CHECK(PA_RV,"ICE4_SLOW end:PA_RV")
  CALL MPPDB_CHECK(PA_RC,"ICE4_SLOW end:PA_RC")
  CALL MPPDB_CHECK(PA_RI,"ICE4_SLOW end:PA_RI")
  CALL MPPDB_CHECK(PA_RS,"ICE4_SLOW end:PA_RS")
  CALL MPPDB_CHECK(PA_RG,"ICE4_SLOW end:PA_RG")
END IF

!$acc end data

!$acc end data

END SUBROUTINE ICE4_SLOW
